% Exact solution of a counting problem from which the significance of the
% Mann-Whitney-Wilcoxon test can be derived.
% Compiling and benchmarking:
% c(erlang_mww).        % for hipe: c(erlang_mww, [native]).
% erlang_mww:timethis(fun() -> erlang_mww:l(28,14,100) end).

-module(erlang_mww).
-export([l/3, distl/3, timethis/1]).

% The quantity l is the number of ways a distribution of N values can be
% divided into two parts of size M and N-M so that Mann-Whitney-U does not
% exceed U.

% Conventional single-threaded recursive calculation of l
l(_, _, U) when U < 0 -> 0;
l(_, M, _) when M =:= 0 -> 1;
l(N, M, U) -> lists:sum([l(N-D-1, M-1, U-M*D) || D <- lists:seq(0, N-M)]).

% Multi-threaded calculation of l spawning one thread for the sum from 1 to the
% maximum at every level of recursion.
distl(_, _, U) when U < 0 -> 0;
distl(_, M, _) when M =:= 0 -> 1;
distl(N, M, U) ->
    Pid= self(),
    spawn(fun() -> lsumfrom1(N, M, U, Pid) end),
    distl(N-1, M-1, U) + (receive X -> X end).

% Compute partial result for l, summing for d from 1 to N at one level of
% recursion.  The result is returned by message passing.
lsumfrom1(N, M, U, P) ->
    P ! lists:sum([l(N-D-1, M-1, U-M*D) || D <- lists:seq(1, N-M)]).

% Output runtime of a 0-ary function.  Returns the function result.
timethis(F) ->
    N0= now(),
    Ret= F(),
    N1= now(),
    N2= now(),
    io:format("~p~n", [2*now2secs(N1) - now2secs(N0) - now2secs(N2)]),
    Ret.

% Convert triple returned by now() to floating-point seconds.
now2secs({M,S,U}) -> 1000000 * M + S + U / 1000000.

